
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

#include "robotis.h"


volatile byte iBuffer[256]; // in interrupt UART read mode, giBuffer[giBufferPointer++] will continously take new read values
volatile byte oBuffer[32]; 
volatile byte iBufferPointer=0;

/*
 * SIGNAL()  Rx Interrupt - write data to buffer (used with CM5's RS485)
 * When a byte (8bit + parity + stop) has been read by the Atmega, it issues an interrupt 
 * RxiBufferPointer goes back to 0 after 255, RxPacket doesn't need it to reset it to 0
 *
 * The reason for using an interrupt driven Rx is because the CM5 has 2 serials.
 * you can't poll two Rx because the other one may need attention during poll.
 */
SIGNAL (SIG_UART0_RECV)
{
	iBuffer[iBufferPointer++] = RXRS485_DATA;
}




/*
 * TxRS485() write byte to Robotis bus 
 */
static void TxRS485(byte Data)
{
	while(!CHECK_TXRS485_READY);
	TXRS485_DATA = Data;
}
/*
 * RxRS485() read byte from Robotis bus (CM5 is normally interrupt Rx so doesnt use this routine)
 */
static byte RxRS485(void)
{

	while(!CHECK_RXRS485_READY); 
	return(RXRS485_DATA);
}


unsigned long Timer(byte mode)
{
	static unsigned long prescaler=0;
	byte h, l;
	
	if (mode) {
		TCCR1B |= mode;
		TCNT1H = 0; // 0 is put in temporary storage
		TCNT1L = 0; // and now both H and L are actually put into TCNT1
					// It took 2 weeks to discover why all kinds of cosmic unexplainables occured
		TCCR1A = 0;
		prescaler = 1<<(3*(mode-1));
		//return (prescaler*((((unsigned long)TCNT1H)<<8) | TCNT1L))>>4;
		return ((((unsigned long)TCNT1H)<<8) | TCNT1L);
	}
	else{
		l= TCNT1L;
		h= TCNT1H;

		//return (prescaler*((((unsigned long)h)<<8) | l))>>4;
		return ((((unsigned long)h)<<8) | l);
	}
	// always returns same units of 1/1,000,000th of a second
	// regardless of prescaler
	//return (prescaler*((((unsigned long)TCNT1H)<<8) | TCNT1L))>>4;
	//return ((((unsigned long)TCNT1H)<<8) | TCNT1L);
}



/*
TxRS485Packet() send Robotis package to RS485.
Similar format packet is sent by master and by slave (5th byte differs)
ID - My ID
Instruction - Master: PING, READ etc. Slave: Error bitmask
ParameterLength - number of parameters sent in iBuffer
Parameter - will be coppied to iBuffer[5]..iBuffer[5+ParameterLength-1] 
*/
void TxRS485Packet(byte ID, byte Instruction, byte ParameterLength, byte *Parameter) 
{
    byte i, CheckSum, PacketLength;

    iBuffer[0] = 0xff;
    iBuffer[1] = 0xff;
    iBuffer[2] = ID;
    iBuffer[3] = ParameterLength+2; //Length(Paramter,Instruction,Checksum)
    iBuffer[4] = Instruction;
	for (i=0; i< ParameterLength;i++)
		iBuffer[5+i] = Parameter[i];

    CheckSum = 0;
    PacketLength = ParameterLength+4+2;
    for(i = 2; i < PacketLength-1; i++) //except 0xff,checksum
    {
        CheckSum += iBuffer[i];
    }
    iBuffer[i] = ~CheckSum; //Writing Checksum with Bit Inversion

    SET_RS485_TXD;
    for(i = 0; i < PacketLength; i++)
    {
        SET_TXRS485_FINISH; // ? this only clears bit6. triggers interrupt if interrupt flag is set
        TxRS485(iBuffer[i]);
    }
    while(!CHECK_TXRS485_FINISH); // Wait until (bit6) TXD Shift register empty
    SET_RS485_RXD;
	
	iBufferPointer=0; // Response packet instantanous. Interrupt Rx relies on buffer being filled from 0
}

/*
 * RxRS485Packet() receive Robotis instruction packet through RS485.
 * This is is the non-interrupt version of the read-packet designed for a sensor board.
 * The full protocol implementation of the protocol is needed only by Robotis servos
 * returns the packet pointer if the packet is intended for me otherwise null
*/
byte *RxRS485Packet(byte ID) 
{
    byte i;
	
PORTD ^= 1<<2;	
    iBuffer[0] = RxRS485(); // 0xff;
    iBuffer[1] = RxRS485(); // 0xff;
	iBuffer[2] = RxRS485(); // ID;
    iBuffer[3] = RxRS485(); // ParameterLength+2; // Length(Paramter,Instruction,Checksum)
    iBuffer[4] = RxRS485(); // Instruction;
	for (i=0; i<iBuffer[3]-2; i++)
		iBuffer[5+i] = RxRS485(); // Parameters..
	iBuffer[5+i] = RxRS485(); // Checksum;
	
	// not performing any error checking. assumning that robotis cables are ok.
	
	return (iBuffer[2] == ID)? iBuffer: (byte *)0;

}



/*
 * RxRS485PacketIntr() receive Robotis (response) packet through RS485.
 * This is is the interrupt-driven version of the read-packet designed for CM5
 * Just waits for the buffer to fill up by the interrupt routine
 *
 * The reason for using an interrupt driven Rx is because the CM5 has 2 serials.
 * you can't poll two Rx because the other one may need attention during poll.
*/

byte *RxRS485PacketIntr(byte ID) 
{
	byte i;
	long r=0;

	Timer(5);

	while (iBufferPointer < 4)
		if (Timer(0)>20) {
			iBufferPointer=0;
			return (byte *)0;
		}

	Timer(5);
	
	while (iBufferPointer < 4+iBuffer[3])
		if (Timer(0)>20) {
			iBufferPointer=0;
			return (byte *)0;
		}

	iBufferPointer=0;
	return (iBuffer[2] == ID)? iBuffer: (byte *)0;

}

/*
 * Robotis_SetTorqueLimit() - sets the torque limits for one or all servos
 * If the torque limit in the Flash memory (P_TORQUE_LIMIT_L) is 0, the servo is in free-run mode
 * limit - 0->0x3ff
 * eeprom_too - sets the eeprom torque limit aswell. It is coppied to FLASH at powerup
 * servo - if 0xfe then broadcast the message to all servos otherwise individual servo
 */

void Robotis_SetTorqueLimit(short limit, byte eeprom_too, byte servo)
{
	// set torque in EEPROM. It is coppied to FLASH at powerup
	oBuffer[0] = P_MAX_TORQUE_L;
	oBuffer[1] = limit&0xff; // P_MAX_TORQUE_L
	oBuffer[2] = (limit&0x300)>>8; // P_MAX_TORQUE_H
	// set torque in FLASH. Coppied from EEPROM at powerup. if 0 then servo is in free-run mode 
	oBuffer[3] = P_TORQUE_LIMIT_L;
	oBuffer[4] = limit&0xff; // P_MAX_TORQUE_L
	oBuffer[5] = (limit&0x300)>>8; // P_MAX_TORQUE_H
	if (servo == 0xfe) {
		TxRS485Packet(0xfe,INST_REG_WRITE,3, oBuffer);
		TxRS485Packet(0xfe,INST_ACTION,0, 0);
		TxRS485Packet(0xfe,INST_REG_WRITE,3, oBuffer+3);
		TxRS485Packet(0xfe,INST_ACTION,0, 0);
	} else {
		TxRS485Packet(servo,INST_WRITE,3, oBuffer);
		RxRS485PacketIntr(servo);
		TxRS485Packet(servo,INST_WRITE,3, oBuffer+3);
		RxRS485PacketIntr(servo);
	}
}

/* Robotis_SetAngleLimit() - set minimum and maximum angles for one or all servos
 * If the maximum angle is 0, the servo goes into free-run mode
 * min - minimum angle
 * max - maximum angle = 0..0x3ff
 * servo - if 0xfe then broadcast to all servos
 */

void Robotis_SetAngleLimit(short min, short max, byte servo)
{
	// set minimum and maxium range values. if max is 0, servo is in free-run mode
	oBuffer[0] = P_CW_ANGLE_LIMIT_L;
	oBuffer[1] = min&0xff; // P_CW_ANGLE_LIMIT_L
	oBuffer[2] = (min&0x300)>>8; // P_CW_ANGLE_LIMIT_H
	oBuffer[3] = max&0xff; // P_CCW_ANGLE_LIMIT_L
	oBuffer[4] = (max&0x300)>>8; // P_CCW_ANGLE_LIMIT_H
	if (servo == 0xfe) {
		TxRS485Packet(0xfe,INST_REG_WRITE,5, oBuffer);
		TxRS485Packet(0xfe,INST_ACTION,0, 0);
	} else {
		TxRS485Packet(servo,INST_WRITE,5, oBuffer);
		RxRS485PacketIntr(servo);
	}
}

/* Robotis_RegWrite() - write similar values to same locations with all servos
 * start_address - in servo's memory
 * nbytes - number of contigous bytes to write to each servo
 * bytes - the byte array to be written
 */

void Robotis_RegWrite(byte start_address, byte nbytes, byte *bytes)
{
	byte *p = oBuffer;
	byte b;
	
	// set minimum and maxium range values. if max is 0, servo is in free-run mode
	*p++ = start_address;
	for (b=0;b<nbytes;b++)
		*p++ = bytes[b];

	TxRS485Packet(0xfe,INST_REG_WRITE,p-oBuffer, oBuffer);
	TxRS485Packet(0xfe,INST_ACTION,0, 0);
}



/* Robotis_SyncWrite() - writes a contigous set of bytes to same location on several servos memory
 * start_address - in servo's memory
 * nbytes - number of bytes to be written to each servo
 * nservos - number of servos to write to
 * bytes - nservos sets of byte arrays
 * servos - list of servo IDs
 */

void Robotis_SyncWrite(byte start_address, byte nbytes, byte nservos, byte *bytes, byte *servos)
{
	byte *p = oBuffer;
	byte s, b;
	
	*p++ = start_address;
	*p++ = nbytes;
	for (s=0;s<nservos;s++) {
		*p++ = servos[s];
		for (b=0;b<nbytes;b++) {
			*p++ = bytes[(s*nbytes)+b];
		}
	}
	TxRS485Packet(0xfe,INST_SYNC_WRITE,p-oBuffer, oBuffer);
}

/* Robotis_ReadMemory() - reads a contigous set of bytes from servo memory
 * servo - ID of servo to read from
 * start_address - in servo's memory
 * nbytes - number of bytes to be read from servo
 */

byte* Robotis_ReadMemory(byte servo, byte start_address, byte nbytes)
{
	byte *p;
	
	oBuffer[0] = start_address; // from the beginning of the servo memory
	oBuffer[1] = nbytes; // read this many bytes
	TxRS485Packet(servo, INST_READ, 2, oBuffer);
	p = RxRS485PacketIntr(servo);
	if (p) return p+5; else return 0;
}

/* Robotis_Ping() - send ping to servo
 * servo - ID of the servo to ping
 * returns 1 if servo responded is there otherwise 0
 */
byte Robotis_Ping(byte servo)
{
	TxRS485Packet(servo, INST_PING, 0, 0);
	return (RxRS485PacketIntr(servo) != 0);
}
